# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["HODLRGP"]

import time

from .basic import GP
from ._george import _george


class HODLRGP(GP):
    """
    A solver built on top of Sivaram Amambikasaran's `HODLR library
    <https://github.com/sivaramambikasaran/HODLR>`_ for linear algebra. The
    HODLR library includes an :math:`\mathcal{O}(N\,\log^2 N)` direct solver
    for dense matrices as described `here <http://arxiv.org/abs/1403.6015>`_.

    :param kernel:
        An instance of a subclass of :class:`kernels.Kernel`.

    :param mean: (optional)
        A description of the mean function; can be a callable or a scalar. If
        scalar, the mean is assumed constant. Otherwise, the function will be
        called with the array of independent coordinates as the only argument.
        (default: ``0.0``)

    :param nleaf: (optional)
        A tuning parameter for the HODLR algorithm. This parameter sets the
        size of the smallest leaf in the tree. (default: ``100``)

    :param tol: (optional)
        A tuning parameter for the HODLR algorithm. This parameter sets the
        low-rank tolerance of the pivoting algorithm. (default: ``1e-12``)

    """

    def __init__(self, kernel, nleaf=100, tol=1e-12, mean=None):
        self.nleaf = nleaf
        self.tol = tol
        self._gp = None
        super(HODLRGP, self).__init__(kernel, mean=mean)

    @property
    def computed(self):
        """
        Has the processes been computed since the last update of the kernel?

        """
        return (self._gp is not None and self._gp.computed()
                and not self.kernel.dirty)

    @property
    def gp(self):
        if self._gp is None or not self.computed:
            self._gp = _george(self.kernel, self.nleaf, self.tol)
        return self._gp

    def _do_compute(self, seed=None):
        if seed is None:
            seed = int(time.time())
        self.gp.compute(self._x, self._yerr, seed)
        self.kernel.dirty = False

    def _compute_lnlike(self, r):
        return self.gp.lnlikelihood(r)

    def grad_lnlikelihood(self, *args, **kwargs):
        raise NotImplementedError("Gradients have not been implemented in the "
                                  "HODLR solver yet.")

    def predict(self, y, t):
        """
        Compute the conditional predictive distribution of the model.

        :param y: ``(nsamples,)``
            The observations to condition the model on.

        :param t: ``(ntest,)`` or ``(ntest, ndim)``
            The coordinates where the predictive distribution should be
            computed.

        Returns a tuple ``(mu, cov)`` where

        * **mu** ``(ntest,)`` is the mean of the predictive distribution, and
        * **cov** ``(ntest, ntest)`` is the predictive covariance.

        """
        r = self._check_dimensions(y)[self.inds] - self.mean(self._x)
        return self.gp.predict(r, self.parse_samples(t, False)[0])

    def get_matrix(self, t):
        """
        Get the covariance matrix at a given set of independent coordinates.

        :param t: ``(nsamples,)`` or ``(nsamples, ndim)``
            The list of samples.

        """
        return self.gp.get_matrix(self.parse_samples(t, False)[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function, absolute_import

__all__ = ["test_basic", "test_predict", "test_sample"]

import numpy as np

from .. import kernels
from ..hodlr import HODLRGP
from ..basic import GP


def test_basic(N=100, seed=123):
    kernel = 1e-6 * kernels.ExpKernel(1.0, 2)
    np.random.seed(seed)
    t = 0.01 * N * np.random.rand(N, kernel.ndim)
    y = np.sin(np.sum(t**2, axis=1))
    yerr = 1e-6 * np.ones_like(y)

    gp1 = GP(kernel)
    gp1.compute(t, yerr)
    ll1 = gp1.lnlikelihood(y)

    gp2 = HODLRGP(kernel)
    gp2.compute(t, yerr)
    ll2 = gp2.lnlikelihood(y)

    assert np.allclose(ll1, ll2), (ll1, ll2)


def test_grad(N=200, seed=123, eps=3.245e-6):
    # Set up the GP.
    kernel = 1e-4 * kernels.ConstantKernel(2.0) \
        + 1e-2 * kernels.ExpSine2Kernel(10.0, 5.0)
    gp = GP(kernel)

    # Generate some fake data.
    np.random.seed(seed)
    t = 10. * np.random.rand(N, kernel.ndim)
    y = gp.sample(t)
    yerr = 0.1 * np.ones_like(y)
    y += yerr * np.random.randn(len(yerr))

    # Compute the analytic gradient.
    gp.compute(t, yerr)
    grad = gp.grad_lnlikelihood(y)

    # Loop over the dimensions and compute the numerical gradient.
    for i in range(len(kernel)):
        kernel[i] += eps
        gp.compute(t, yerr)
        llp = gp.lnlikelihood(y)

        kernel[i] -= 2 * eps
        gp.compute(t, yerr)
        llm = gp.lnlikelihood(y)
        g = 0.5 * (llp - llm) / eps

        assert np.allclose(g, grad[i]), \
            "Gradient computation failed in dimension {0}".format(i)
        kernel[i] += eps


def test_predict(N=20, N2=50, seed=123):
    kernel = 1e-6 * kernels.ExpKernel(1.0, 2)
    np.random.seed(seed)

    t = np.random.randn(N, kernel.ndim)
    t0 = np.random.randn(N2, kernel.ndim)

    y = np.sin(np.sum(t**2, axis=1))
    yerr = 1e-6 * np.ones_like(y)

    gp1 = GP(kernel)
    gp1.compute(t, yerr)
    mu1, cov1 = gp1.predict(y, t0)

    gp2 = HODLRGP(kernel)
    gp2.compute(t, yerr)
    mu2, cov2 = gp2.predict(y, t0)

    assert np.allclose(mu1, mu2), "The predictive means don't match"
    assert np.allclose(cov1, cov2), \
        "The predictive covariances don't match"


def test_sample(N=20, N2=50, seed=123):
    kernel = 1e-6 * kernels.ExpKernel(1.0, 2)
    np.random.seed(seed)

    t = np.random.randn(N, kernel.ndim)
    t0 = np.random.randn(N2, kernel.ndim)

    y = np.sin(np.sum(t**2, axis=1))
    yerr = 1e-6 * np.ones_like(y)

    gp1 = GP(kernel)
    gp1.compute(t, yerr)

    np.random.seed(seed)
    samps1 = gp1.sample_conditional(y, t0, 6)

    gp2 = HODLRGP(kernel)
    gp2.compute(t, yerr)

    np.random.seed(seed)
    samps2 = gp2.sample_conditional(y, t0, 6)

    assert np.allclose(samps1, samps2)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
    "Sum", "Product", "Kernel",
    "ConstantKernel", "WhiteKernel", "DotProductKernel",
    "RadialKernel", "ExpKernel", "ExpSquaredKernel", "RBFKernel",
    "CosineKernel", "ExpSine2Kernel",
    "Matern32Kernel", "Matern52Kernel",
]

import numpy as np
from scipy.linalg import cho_factor, cho_solve


class Kernel(object):
    """
    The abstract kernel type. Every kernel implemented in George should be
    a subclass of this object.

    :param pars:
        The hyper-parameters of the kernel.

    :param ndim: (optional)
        The number of input dimensions of the kernel. (default: ``1``)

    """

    is_kernel = True
    kernel_type = -1

    def __init__(self, *pars, **kwargs):
        self.ndim = kwargs.get("ndim", 1)
        self.pars = np.array(pars)

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__,
                                 ", ".join(map("{0}".format,
                                               self.pars) +
                                           ["ndim={0}".format(self.ndim)]))

    @property
    def pars(self):
        """
        A NumPy array listing the vector of hyperparameters specifying the
        kernel.

        """
        return self._pars

    @pars.setter
    def pars(self, v):
        self._pars = np.array(v)
        self.set_pars(v)
        self.dirty = True

    def __getitem__(self, i):
        return self._pars[i]

    def __setitem__(self, i, v):
        self._pars[i] = v
        self.pars = self._pars

    def set_pars(self, pars):
        pass

    def __len__(self):
        return len(self.pars)

    def __add__(self, b):
        if not hasattr(b, "is_kernel"):
            return Sum(ConstantKernel(np.sqrt(np.abs(float(b))),
                                      ndim=self.ndim), self)
        return Sum(self, b)

    def __radd__(self, b):
        return self.__add__(b)

    def __mul__(self, b):
        if not hasattr(b, "is_kernel"):
            return Product(ConstantKernel(np.sqrt(np.abs(float(b))),
                                          ndim=self.ndim), self)
        return Product(self, b)

    def __rmul__(self, b):
        return self.__mul__(b)

    def __call__(self, x1, x2):
        """
        The value of the kernel evaluated at a specific pair of coordinates.

        """
        raise NotImplementedError("Must be implemented by subclasses")

    def grad(self, x1, x2):
        """
        The kernel gradient evaluated at a specific pair of coordinates. The
        order of the parameters is the same as the ``.pars`` property.

        """
        raise NotImplementedError("Must be implemented by subclasses")


class _operator(Kernel):
    is_kernel = False
    operator_type = -1

    def __init__(self, k1, k2):
        if k1.ndim != k2.ndim:
            raise ValueError("Dimension mismatch")
        self.k1 = k1
        self.k2 = k2
        self.ndim = k1.ndim
        self.dirty = True

    @property
    def pars(self):
        return np.append(self.k1.pars, self.k2.pars)

    @pars.setter
    def pars(self, v):
        self.dirty = True
        i = len(self.k1)
        self.k1.pars = v[:i]
        self.k2.pars = v[i:]

    def __getitem__(self, i):
        return self.pars[i]

    def __setitem__(self, i, v):
        p = self.pars
        p[i] = v
        self.pars = p


class Sum(_operator):
    is_kernel = False
    operator_type = 0

    def __repr__(self):
        return "{0} + {1}".format(self.k1, self.k2)

    def __call__(self, x1, x2):
        return self.k1(x1, x2) + self.k2(x1, x2)

    def grad(self, x1, x2):
        g1 = self.k1.grad(x1, x2)
        g2 = self.k2.grad(x1, x2)
        return np.concatenate((g1, g2), axis=0)


class Product(_operator):
    is_kernel = False
    operator_type = 1

    def __repr__(self):
        return "{0} * {1}".format(self.k1, self.k2)

    def __call__(self, x1, x2):
        return self.k1(x1, x2) * self.k2(x1, x2)

    def grad(self, x1, x2):
        g1 = self.k1.grad(x1, x2)
        g2 = self.k2.grad(x1, x2)
        s = [None] + [slice(None)] * (len(g1.shape) - 1)
        return np.concatenate((g1 * self.k2(x1, x2)[s],
                               g2 * self.k1(x1, x2)[s]), axis=0)


class ConstantKernel(Kernel):
    r"""
    This kernel returns the constant

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = c^2

    where :math:`c` is a parameter.

    :param value:
        The constant value :math:`c` in the above equation.

    """
    kernel_type = 0

    def __init__(self, value, ndim=1):
        super(ConstantKernel, self).__init__(value, ndim=ndim)

    def __call__(self, x1, x2):
        return self.pars[0] ** 2 + np.sum(np.zeros_like(x1 - x2), axis=-1)

    def grad(self, x1, x2):
        x = np.sum(np.zeros_like(x1 - x2), axis=-1)
        return 2 * self.pars[0] + np.zeros(np.append(1, x.shape))


class WhiteKernel(Kernel):
    r"""
    This kernel returns constant along the diagonal.

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = c^2 \, \delta_{ij}

    where :math:`c` is the parameter.

    :param value:
        The constant value :math:`c` in the above equation.

    """
    kernel_type = 8

    def __init__(self, value, ndim=1):
        super(WhiteKernel, self).__init__(value, ndim=ndim)

    def __call__(self, x1, x2):
        d = np.sum((x1 - x2) ** 2, axis=-1)
        return self.pars[0] ** 2 * (d == 0.0)

    def grad(self, x1, x2):
        d = np.sum((x1 - x2) ** 2, axis=-1)
        g = np.zeros(np.append(1, d.shape))
        g[0] = 2 * self.pars[0] * (d == 0)
        return g


class DotProductKernel(Kernel):
    r"""
    The dot-product kernel takes the form

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = \mathbf{x}_i^{\mathrm{T}} \cdot
                                         \mathbf{x}_j

    """
    kernel_type = 1

    def __init__(self, ndim=1):
        super(DotProductKernel, self).__init__(ndim=ndim)

    def __call__(self, x1, x2):
        return np.sum(x1 * x2, axis=-1)

    def grad(self, x1, x2):
        return np.zeros(np.append(1, self(x1, x2).shape))


class RadialKernel(Kernel):
    r"""
    This abstract base class implements a radial kernel in an arbitrary
    metric.  The metric is specified as a matrix :math:`C` where the
    radius :math:`{r_{ij}}^2` is

    .. math::

        {r_{ij}}^2 = (\mathbf{x}_i - \mathbf{x}_j)^\mathrm{T}\,
                     C^{-1}\,(\mathbf{x}_i - \mathbf{x}_j)

    :param metric:
        There are a few different ways that you can specify the metric:

        1. if ``metric`` is a scalar, the metric is assumed isotropic with an
           axis-aligned variance of ``metric`` in each dimension,
        2. if ``metric`` is one-dimensional, it is assumed to specify the
           axis-aligned variances in each dimension, and
        3. if ``metric`` is two-dimensional, it is assumed to give the full
           matrix :math:`C`.

    However you specify ``metric``, the ``pars`` property of the kernel will
    be a NumPy array with elements:

    .. code-block:: python

        kernel.pars = [a, b, c, d, e, f, ...]

    where

    .. math::

        C = \left ( \begin{array}{cccc}
            a & b & d & \\
            b & c & e & \cdots \\
            d & e & f & \\
             & \vdots & & \ddots
        \end{array} \right )

    **Note:**
    Subclasses should implement the :func:`get_value` method to give
    the value of the kernel at a given radius and this class will deal with
    the metric.

    **Another Note:**
    1-D is treated as a special case for speed.

    """

    def __init__(self, metric, ndim=1, extra=[]):
        # Special case 1-D for speed.
        if ndim == 1:
            self.nextra = len(extra)
            super(RadialKernel, self).__init__(*(np.append(extra, metric)),
                                               ndim=1)
        else:
            inds = np.tri(ndim, dtype=bool)
            try:
                l = len(metric)
            except TypeError:
                pars = np.diag(float(metric) * np.ones(ndim))[inds]
            else:
                if l == ndim:
                    pars = np.diag(metric)[inds]
                else:
                    pars = np.array(metric)
                    if l != (ndim*ndim + ndim) / 2:
                        raise ValueError("Dimension mismatch")
            self.nextra = len(extra)
            super(RadialKernel, self).__init__(*(np.append(extra, pars)),
                                               ndim=ndim)

            # Build the gradient indicator masks.
            self.gm = np.empty(np.append(len(pars), self.matrix.shape),
                               dtype=int)
            for i in range(len(pars)):
                ind = np.zeros(len(pars), dtype=int)
                ind[i] = 1
                self.gm[i] = self._build_matrix(ind)

    def _build_matrix(self, x, dtype=float):
        m = np.zeros((self.ndim, self.ndim), dtype=dtype)
        m[np.tril_indices_from(m)] = x
        m += m.T
        m[np.diag_indices_from(m)] *= 0.5
        return m

    def set_pars(self, pars):
        if self.ndim > 1:
            self.matrix = self._build_matrix(pars[self.nextra:])
            self._factor = cho_factor(self.matrix)
        else:
            self.ivar = 1.0 / pars[self.nextra]

    def __call__(self, x1, x2):
        if self.ndim > 1:
            dx = x1 - x2
            dxf = dx.reshape((-1, self.ndim)).T
            r = np.sum(dxf * cho_solve(self._factor, dxf), axis=0)
            r = r.reshape(dx.shape[:-1])
        else:
            r = np.sum((x1 - x2), axis=-1) ** 2 * self.ivar
        return self.get_value(r)

    def grad(self, x1, x2):
        if self.ndim > 1:
            dx = x1 - x2
            dxf = dx.reshape((-1, self.ndim)).T
            alpha = cho_solve(self._factor, dxf)

            # Compute the radial gradient.
            g = np.empty(np.append(len(self.gm)+self.nextra, dx.shape[:-1]))
            for i, gm in enumerate(self.gm):
                g[self.nextra+i] = \
                    np.sum(dxf*cho_solve(self._factor, np.dot(gm, alpha)),
                           axis=0).reshape(dx.shape[:-1])

            # Compute the function gradient.
            r = np.sum(dxf * alpha, axis=0)
            r = r.reshape(dx.shape[:-1])
            if self.nextra:
                kg, g[:self.nextra] = self.get_grad(r)
            else:
                kg = self.get_grad(r)

            # Update the full gradient to include the kernel.
            s = [None] + [slice(None)] * len(r.shape)
            g[self.nextra:] *= -kg[s]

        else:
            r = np.sum((x1 - x2), axis=-1) ** 2 * self.ivar

            # Compute the radial gradient.
            g = np.empty(np.append(len(self), r.shape))
            if self.nextra:
                kg, g[:self.nextra] = self.get_grad(r)
            else:
                kg = self.get_grad(r)

            g[self.nextra] = -kg * r * self.ivar

        return g

    def get_value(self, r):
        raise NotImplementedError("Subclasses must implement this method.")

    def get_grad(self, r):
        raise NotImplementedError("Subclasses must implement this method.")


class ExpKernel(RadialKernel):
    r"""
    The exponential kernel is a :class:`RadialKernel` where the value at a
    given radius :math:`r^2` is given by:

    .. math::

        k({r_{ij}}) = \exp \left ( -|r| \right )

    :param metric:
        The custom metric specified as described in the :class:`RadialKernel`
        description.

    """
    kernel_type = 2

    def get_value(self, dx):
        return np.exp(-np.sqrt(dx))

    def get_grad(self, dx):
        sx = np.sqrt(dx)
        m = sx > 0
        g = np.zeros_like(sx)
        g[m] = -0.5 * np.exp(-sx[m]) / sx[m]
        return g


class ExpSquaredKernel(RadialKernel):
    r"""
    The exponential-squared kernel is a :class:`RadialKernel` where the value
    at a given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \exp \left ( -\frac{r^2}{2} \right )

    :param metric:
        The custom metric specified as described in the :class:`RadialKernel`
        description.

    """
    kernel_type = 3

    def get_value(self, dx):
        return np.exp(-0.5 * dx)

    def get_grad(self, dx):
        return -0.5 * np.exp(-0.5 * dx)


class RBFKernel(ExpSquaredKernel):
    r"""
    An alias for :class:`ExpSquaredKernel`.

    """


class Matern32Kernel(RadialKernel):
    r"""
    The Matern-3/2 kernel is a :class:`RadialKernel` where the value at a
    given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \left( 1+\sqrt{3\,r^2} \right)\,
                 \exp \left (-\sqrt{3\,r^2} \right )

    :param metric:
        The custom metric specified as described in the :class:`RadialKernel`
        description.

    """
    kernel_type = 6

    def get_value(self, dx):
        r = np.sqrt(3.0 * dx)
        return (1.0 + r) * np.exp(-r)

    def get_grad(self, dx):
        r = np.sqrt(3.0 * dx)
        return -3 * 0.5 * np.exp(-r)


class Matern52Kernel(RadialKernel):
    r"""
    The Matern-5/2 kernel is a :class:`RadialKernel` where the value at a
    given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \left( 1+\sqrt{5\,r^2} + \frac{5\,r^2}{3} \right)\,
                 \exp \left (-\sqrt{5\,r^2} \right )

    :param metric:
        The custom metric specified as described in the :class:`RadialKernel`
        description.

    """
    kernel_type = 7

    def get_value(self, dx):
        r = np.sqrt(5.0 * dx)
        return (1.0 + r + r*r / 3.0) * np.exp(-r)

    def get_grad(self, dx):
        r = np.sqrt(5.0 * dx)
        return -5 * (1.0 + r) * np.exp(-r) / 6.0


class RationalQuadraticKernel(RadialKernel):
    r"""
    The Matern-5/2 kernel is a :class:`RadialKernel` where the value at a
    given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \left( 1+ \frac{r^2}{2\,\alpha} \right )^{-\alpha}

    :param alpha:
        The shape parameter :math:`\alpha`.

    :param metric:
        The custom metric specified as described in the :class:`RadialKernel`
        description.

    """
    kernel_type = 9

    def __init__(self, alpha, metric, ndim=1):
        super(RationalQuadraticKernel, self).__init__(metric, extra=[alpha],
                                                      ndim=ndim)

    def get_value(self, dx):
        a = self.pars[0]
        return (1.0 + 0.5 * dx / a) ** (-a)

    def get_grad(self, dx):
        a = self.pars[0]
        t1 = 1 + 0.5 * dx / a
        t2 = 2 * a + dx
        return -0.5 * t1**(-a-1), t1 ** (-a) * (dx - t2 * np.log(t1)) / t2


class CosineKernel(Kernel):
    r"""
    The cosine kernel is given by:

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) =
            \cos\left(\frac{2\,\pi}{P}\,\left|x_i-x_j\right| \right)

    where :math:`P` is the period.

    :param period:
        The period :math:`P` of the oscillation (in the same units as
        :math:`\mathbf{x}`).

    **Note:**
    A shortcoming of this kernel is that it currently only accepts a single
    period so it's not very applicable to problems with input dimension larger
    than one.

    """
    kernel_type = 4

    def __init__(self, period, ndim=1):
        super(CosineKernel, self).__init__(period, ndim=ndim)

    def set_pars(self, pars, twopi=2*np.pi):
        self._omega = twopi / float(pars)

    def __call__(self, x1, x2):
        return np.cos(self._omega * np.sqrt(np.sum((x1 - x2) ** 2, axis=-1)))

    def grad(self, x1, x2, itwopi=1.0/(2*np.pi)):
        x = np.sqrt(np.sum((x1 - x2) ** 2, axis=-1))
        g = np.empty(np.append(1, x.shape))
        s = [0] + [slice(None)] * len(x.shape)
        g[s] = itwopi * x * np.sin(self._omega * x) * self._omega ** 2
        return g


class ExpSine2Kernel(Kernel):
    r"""
    The exp-sine-squared kernel is used to model stellar rotation and *might*
    be applicable in some other contexts. It is given by the equation:

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) =
            \sin \left( -\Gamma\,\sin^2\left[
                \frac{\pi}{P}\,\left|x_i-x_j\right|
            \right] \right)

    where :math:`\Gamma` is the "scale" of the correlation and :math:`P` is
    the period of the oscillation measured in the same units as
    :math:`\mathbf{x}`.

    :param gamma:
        The scale :math:`\Gamma` of the correlations.

    :param period:
        The period :math:`P` of the oscillation (in the same units as
        :math:`\mathbf{x}`).

    **Note:**
    A shortcoming of this kernel is that it currently only accepts a single
    period and scale so it's not very applicable to problems with input
    dimension larger than one.

    """
    kernel_type = 5

    def __init__(self, gamma, period, ndim=1):
        super(ExpSine2Kernel, self).__init__(gamma, period, ndim=ndim)

    def set_pars(self, pars):
        self._omega = np.pi / pars[1]

    def __call__(self, x1, x2):
        s = np.sin(self._omega * np.sqrt(np.sum((x1 - x2) ** 2, axis=-1)))
        return np.exp(-self.pars[0] * s**2)

    def grad(self, x1, x2):
        # Pre-compute some factors.
        x = np.sqrt(np.sum((x1 - x2) ** 2, axis=-1))
        sx = np.sin(self._omega * x)
        cx = np.cos(self._omega * x)
        A2 = sx*sx
        a = self.pars[0]
        f = np.exp(-a * A2)

        # Build the output array.
        g = np.empty(np.append(2, x.shape))
        s = [0] + [slice(None)] * len(x.shape)

        # Compute the scale derivative.
        g[s] = -f * A2

        # Compute the period derivative.
        s[0] = 1
        g[s] = 2 * f * a * sx * cx * x * self._omega ** 2 / np.pi

        return g



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function, absolute_import

__all__ = [
    "test_constant", "test_white", "test_dot_prod",

    "test_exp", "test_exp_squared", "test_rbf", "test_matern32",
    "test_matern52", "test_rational_quadratic",

    "test_cosine", "test_exp_sine2",

    "test_combine",
]

import numpy as np

from .. import kernels
from ..hodlr import HODLRGP


def do_kernel_t(kernel, N=20, seed=123, eps=1.32e-7):
    """
    Test that both the Python and C++ kernels return the same matrices.

    """
    np.random.seed(seed)
    t = np.random.randn(N, kernel.ndim)
    gp = HODLRGP(kernel)

    # Compute the two matrices.
    k1 = kernel(t[:, None], t[None, :])
    k2 = gp.get_matrix(t)

    # Build the matrix using brute force and check that.
    for i, a in enumerate(t):
        for j, b in enumerate(t):
            v = kernel(a, b)
            assert np.allclose(k1[i, j], v), \
                "Python matrix fails (ndim = {2})\n{0} should be {1}" \
                .format(k1[i, j], v, kernel.ndim)
            assert np.allclose(k2[i, j], v), \
                "C++ matrix fails (ndim = {2})\n{0} should be {1}" \
                .format(k2[i, j], v, kernel.ndim)

    # Test that C++ and Python give the same matrix.
    assert np.allclose(k1, k2), (k1, k2)

    # Check the gradients.
    g1 = kernel.grad(t[:, None], t[None, :])
    for i in range(len(kernel)):
        # Compute the centered finite difference approximation to the gradient.
        kernel[i] += eps
        kp = kernel(t[:, None], t[None, :])
        kernel[i] -= 2*eps
        km = kernel(t[:, None], t[None, :])
        kernel[i] += eps
        g0 = 0.5 * (kp - km) / eps
        assert np.allclose(g1[i], g0), \
            "Python gradient computation failed in dimension {0}".format(i)


#
# BASIC KERNELS
#

def test_constant():
    do_kernel_t(kernels.ConstantKernel(0.1))
    do_kernel_t(kernels.ConstantKernel(10.0, 2))
    do_kernel_t(kernels.ConstantKernel(5.0, 5))


def test_white():
    do_kernel_t(kernels.WhiteKernel(0.1))
    do_kernel_t(kernels.WhiteKernel(10.0, 2))
    do_kernel_t(kernels.WhiteKernel(5.0, 5))


def test_dot_prod():
    do_kernel_t(kernels.DotProductKernel())
    do_kernel_t(kernels.DotProductKernel(2))
    do_kernel_t(kernels.DotProductKernel(5))


#
# COVARIANCE KERNELS
#

def do_cov_t(kernel_type, extra=None):
    def build_kernel(metric, **kwargs):
        if extra is None:
            return kernel_type(metric, **kwargs)
        return kernel_type(*(extra + [metric]), **kwargs)

    kernel = build_kernel(0.1)
    do_kernel_t(kernel)

    kernel = build_kernel(1.0)
    do_kernel_t(kernel)

    kernel = build_kernel(10.0)
    do_kernel_t(kernel)

    m = [1.0,
         0.5, 2.0,
         0.1, 0.3, 0.7]
    kernel = build_kernel(m, ndim=3)
    assert np.allclose(kernel.matrix, [[1.0, 0.5, 0.1],
                                       [0.5, 2.0, 0.3],
                                       [0.1, 0.3, 0.7]])
    do_kernel_t(kernel)

    kernel = build_kernel([1.0, 0.1, 10.0], ndim=3)
    do_kernel_t(kernel)

    kernel = build_kernel(1.0, ndim=3)
    do_kernel_t(kernel)

    try:
        kernel = build_kernel([1.0, 0.1, 10.0, 500], ndim=3)
    except ValueError:
        pass
    else:
        assert False, "This test should fail"


def test_exp():
    do_cov_t(kernels.ExpKernel)


def test_exp_squared():
    do_cov_t(kernels.ExpSquaredKernel)


def test_rbf():
    do_cov_t(kernels.RBFKernel)


def test_matern32():
    do_cov_t(kernels.Matern32Kernel)


def test_matern52():
    do_cov_t(kernels.Matern52Kernel)


def test_rational_quadratic():
    do_cov_t(kernels.RationalQuadraticKernel, [1.0])
    do_cov_t(kernels.RationalQuadraticKernel, [0.1])
    do_cov_t(kernels.RationalQuadraticKernel, [10.0])


#
# PERIODIC KERNELS
#

def test_cosine():
    do_kernel_t(kernels.CosineKernel(1.0))
    do_kernel_t(kernels.CosineKernel(0.5, 2))
    do_kernel_t(kernels.CosineKernel(0.75, 5))


def test_exp_sine2():
    do_kernel_t(kernels.ExpSine2Kernel(0.4, 1.0))
    do_kernel_t(kernels.ExpSine2Kernel(12.0, 0.5, 2))
    do_kernel_t(kernels.ExpSine2Kernel(13.7, 0.75, 5))


#
# COMBINING KERNELS
#

def test_combine():
    do_kernel_t(12 * kernels.ExpSine2Kernel(0.4, 1.0, ndim=5) + 0.1)
    do_kernel_t(12 * kernels.ExpSquaredKernel(0.4, ndim=3) + 0.1)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["multivariate_gaussian_samples", "nd_sort_samples"]

import numpy as np
from scipy.spatial import cKDTree


def multivariate_gaussian_samples(matrix, N, mean=None):
    """
    Generate samples from a multidimensional Gaussian with a given covariance.

    :param matrix: ``(k, k)``
        The covariance matrix.

    :param N:
        The number of samples to generate.

    :param mean: ``(k,)`` (optional)
        The mean of the Gaussian. Assumed to be zero if not given.

    :returns samples: ``(k,)`` or ``(N, k)``
        Samples from the given multivariate normal.

    """
    if mean is None:
        mean = np.zeros(len(matrix))
    samples = np.random.multivariate_normal(mean, matrix, N)
    if N == 1:
        return samples[0]
    return samples


def nd_sort_samples(samples):
    """
    Sort an N-dimensional list of samples using a KDTree.

    :param samples: ``(nsamples, ndim)``
        The list of samples. This must be a two-dimensional array.

    :returns i: ``(nsamples,)``
        The list of indices into the original array that return the correctly
        sorted version.

    """
    # Check the shape of the sample list.
    assert len(samples.shape) == 2

    # Build a KD-tree on the samples.
    tree = cKDTree(samples)

    # Compute the distances.
    d, i = tree.query(samples[0], k=len(samples))
    return i



#!/usr/bin/env python

import os
import re

try:
    from setuptools import setup, Extension
    from setuptools.command.build_ext import build_ext as _build_ext
except ImportError:
    from distutils.core import setup, Extension
    from distutils.command.build_ext import build_ext as _build_ext


def find_eigen(hint=None):
    """
    Find the location of the Eigen 3 include directory. This will return
    ``None`` on failure.

    """
    # List the standard locations including a user supplied hint.
    search_dirs = [] if hint is None else hint
    search_dirs += [
        "/usr/local/include/eigen3",
        "/usr/local/homebrew/include/eigen3",
        "/opt/local/var/macports/software/eigen3",
        "/opt/local/include/eigen3",
        "/usr/include/eigen3",
        "/usr/include/local",
    ]

    # Loop over search paths and check for the existence of the Eigen/Dense
    # header.
    for d in search_dirs:
        path = os.path.join(d, "Eigen", "Dense")
        if os.path.exists(path):
            # Determine the version.
            vf = os.path.join(d, "Eigen", "src", "Core", "util", "Macros.h")
            if not os.path.exists(vf):
                continue
            src = open(vf, "r").read()
            v1 = re.findall("#define EIGEN_WORLD_VERSION (.+)", src)
            v2 = re.findall("#define EIGEN_MAJOR_VERSION (.+)", src)
            v3 = re.findall("#define EIGEN_MINOR_VERSION (.+)", src)
            if not len(v1) or not len(v2) or not len(v3):
                continue
            v = "{0}.{1}.{2}".format(v1[0], v2[0], v3[0])
            print("Found Eigen version {0} in: {1}".format(v, d))
            return d
    return None


def find_hodlr(hint=None):
    """
    Find the location of the HODLR include directory. This will return
    ``None`` on failure.

    """
    # List the standard locations including a user supplied hint.
    search_dirs = [] if hint is None else hint
    search_dirs += [
        "./hodlr/header",
        "/usr/local/include",
        "/usr/include/local",
    ]

    # Loop over search paths and check for the existence of the Eigen/Dense
    # header.
    for d in search_dirs:
        paths = [
            os.path.join(d, "HODLR_Matrix.hpp"),
            os.path.join(d, "HODLR_Tree.hpp"),
            os.path.join(d, "HODLR_Node.hpp"),
        ]
        if all(map(os.path.exists, paths)):
            print("Found HODLR headers in: {0}".format(d))
            return d
    return None


class build_ext(_build_ext):
    """
    A custom extension builder that finds the include directories for Eigen
    and HODLR before compiling.

    """

    def build_extension(self, ext):
        dirs = ext.include_dirs + self.compiler.include_dirs

        # Look for the Eigen headers and make sure that we can find them.
        eigen_include = find_eigen(hint=dirs)
        if eigen_include is None:
            raise RuntimeError("Required library Eigen 3 not found. "
                               "Check the documentation for solutions.")

        # Look for the HODLR headers and make sure that we can find them.
        hodlr_include = find_hodlr(hint=dirs)
        if hodlr_include is None:
            raise RuntimeError("Required library HODLR not found. "
                               "Check the documentation for solutions.")

        # Update the extension's include directories.
        ext.include_dirs += [eigen_include, hodlr_include]

        # Run the standard build procedure.
        _build_ext.build_extension(self, ext)


if __name__ == "__main__":
    import sys
    import numpy

    # Publish the library to PyPI.
    if "publish" in sys.argv[-1]:
        os.system("python setup.py sdist upload")
        sys.exit()

    # Set up the C++-extension.
    libraries = []
    if os.name == "posix":
        libraries.append("m")
    include_dirs = [
        "include",
        numpy.get_include(),
    ]
    ext = Extension("george._george", sources=["george/_george.cc"],
                    libraries=libraries, include_dirs=include_dirs)

    # Hackishly inject a constant into builtins to enable importing of the
    # package before the library is built.
    if sys.version_info[0] < 3:
        import __builtin__ as builtins
    else:
        import builtins
    builtins.__GEORGE_SETUP__ = True
    import george

    setup(
        name="george",
        version=george.__version__,
        author="Daniel Foreman-Mackey",
        author_email="danfm@nyu.edu",
        url="https://github.com/dfm/george",
        license="MIT",
        packages=["george", "george.testing"],
        ext_modules=[ext],
        description="Blazingly fast Gaussian Processes for regression.",
        long_description=open("README.rst").read(),
        package_data={"": ["README.rst", "LICENSE",
                           "include/*.h", "hodlr/header/*.hpp", ]},
        include_package_data=True,
        cmdclass=dict(build_ext=build_ext),
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
        ],
    )



# -*- coding: utf-8 -*-

__version__ = "0.1.0"

try:
    __GEORGE_SETUP__
except NameError:
    __GEORGE_SETUP__ = False

if not __GEORGE_SETUP__:
    __all__ = ["kernels", "GP", "HODLRGP"]

    from . import kernels
    from .basic import GP
    from .hodlr import HODLRGP

    # Tests.
    from .testing.test_basic import *  # NOQA
    from .testing.test_kernels import *  # NOQA



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["GP"]

try:
    from itertools import izip
except ImportError:
    izip = zip

import numpy as np
import scipy.optimize as op
from scipy.linalg import cho_factor, cho_solve, LinAlgError

from .utils import multivariate_gaussian_samples, nd_sort_samples


# MAGIC: tiny epsilon to add on the diagonal of the matrices in the absence
# of observational uncertainties. Needed for computational stability.
TINY = 1.25e-12


class GP(object):
    """
    The basic Gaussian Process object.

    :param kernel:
        An instance of a subclass of :class:`kernels.Kernel`.

    :param mean: (optional)
        A description of the mean function; can be a callable or a scalar. If
        scalar, the mean is assumed constant. Otherwise, the function will be
        called with the array of independent coordinates as the only argument.
        (default: ``0.0``)

    """

    def __init__(self, kernel, mean=None):
        self.kernel = kernel
        self._computed = False
        if mean is None:
            self.mean = _default_mean(0.)
        else:
            try:
                val = float(mean)
            except TypeError:
                self.mean = mean
            else:
                self.mean = _default_mean(val)

    @property
    def computed(self):
        """
        Has the processes been computed since the last update of the kernel?

        """
        return self._computed and not self.kernel.dirty

    @computed.setter
    def computed(self, v):
        self._computed = v
        if v:
            self.kernel.dirty = False

    def parse_samples(self, t, sort=False):
        """
        Parse a list of samples to make sure that it has the correct
        dimensions and optionally sort it. In one dimension, the samples will
        be sorted in the logical order. In higher dimensions, a kd-tree is
        built and the samples are sorted in increasing distance from the
        *first* sample.

        :param t: ``(nsamples,)`` or ``(nsamples, ndim)``
            The list of samples. If 1-D, this is assumed to be a list of
            one-dimensional samples otherwise, the size of the second
            dimension is assumed to be the dimension of the input space.

        :param sort:
            A boolean flag indicating whether or not the samples should be
            sorted.

        Returns a tuple ``(samples, inds)`` where

        * **samples** is an array with shape ``(nsamples, ndim)`` and if
          ``sort`` was ``True``, it will also be sorted, and
        * **inds** is an ``(nsamples,)`` list of integer permutations used to
          sort the list of samples.

        Raises a ``RuntimeError`` if the input dimension doesn't match the
        dimension of the kernel.

        """
        t = np.atleast_1d(t)
        if len(t.shape) == 1:
            # Deal with one-dimensional data.
            if sort:
                inds = np.argsort(t)
            else:
                inds = np.arange(len(t), dtype=int)
            t = np.atleast_2d(t).T
        elif sort:
            # Sort the data using a KD-tree.
            inds = nd_sort_samples(t)
        else:
            # Otherwise, assume that the samples are sorted.
            inds = np.arange(t.shape[0], dtype=int)

        # Double check the dimensions against the kernel.
        if len(t.shape) != 2 or t.shape[1] != self.kernel.ndim:
            raise ValueError("Dimension mismatch")

        return t[inds], inds

    def _check_dimensions(self, y):
        n, ndim = self._x.shape
        y = np.atleast_1d(y)
        if len(y.shape) > 1:
            raise ValueError("The predicted dimension must be 1-D")
        if len(y) != n:
            raise ValueError("Dimension mismatch")
        return y

    def compute(self, x, yerr=TINY, sort=True, **kwargs):
        """
        Pre-compute the covariance matrix and factorize it for a set of times
        and uncertainties.

        :param x: ``(nsamples,)`` or ``(nsamples, ndim)``
            The independent coordinates of the data points.

        :param yerr: (optional) ``(nsamples,)`` or scalar
            The Gaussian uncertainties on the data points at coordinates
            ``x``. These values will be added in quadrature to the diagonal of
            the covariance matrix.

        :param sort: (optional)
            Should the samples be sorted before computing the covariance
            matrix? This can lead to more numerically stable results and with
            some linear algebra libraries this can more computationally
            efficient. Either way, this flag is passed directly to
            :func:`parse_samples`. (default: ``True``)

        """
        # Parse the input coordinates.
        self._x, self.inds = self.parse_samples(x, sort)
        try:
            self._yerr = float(yerr) * np.ones(len(x))
        except TypeError:
            self._yerr = self._check_dimensions(yerr)[self.inds]
        self._do_compute(**kwargs)

    def _do_compute(self, _scale=0.5*np.log(2*np.pi)):
        # Compute the kernel matrix.
        K = self.kernel(self._x[:, None], self._x[None, :])
        K[np.diag_indices_from(K)] += self._yerr ** 2

        # Factor the matrix and compute the log-determinant.
        factor, _ = self._factor = cho_factor(K, overwrite_a=True)
        self._const = -(np.sum(np.log(np.diag(factor))) + _scale*len(self._x))

        # Save the computed state.
        self.computed = True

    def recompute(self, sort=False, **kwargs):
        """
        Re-compute a previously computed model. You might want to do this if
        the kernel parameters change and the kernel is labeled as ``dirty``.

        :params sort: (optional)
            Should the samples be sorted before computing the covariance
            matrix? (default: ``False``)

        """
        if not (hasattr(self, "_x") and hasattr(self, "_yerr")):
            raise RuntimeError("You need to compute the model first")
        return self.compute(self._x, self._yerr, sort=sort, **kwargs)

    def _compute_lnlike(self, r):
        return self._const - 0.5*np.dot(r.T, cho_solve(self._factor, r))

    def lnlikelihood(self, y, quiet=False):
        """
        Compute the ln-likelihood of a set of observations under the Gaussian
        process model. You must call ``compute`` before this function.

        :param y: ``(nsamples, )``
            The observations at the coordinates provided in the ``compute``
            step.

        :param quiet:
            If ``True`` return negative infinity instead of raising an
            exception when there is an invalid kernel or linear algebra
            failure. (default: ``False``)

        """
        if not self.computed:
            try:
                self.recompute()
            except (ValueError, LinAlgError):
                if quiet:
                    return -np.inf
                raise
        r = self._check_dimensions(y)[self.inds] - self.mean(self._x)
        ll = self._compute_lnlike(r)
        return ll if np.isfinite(ll) else -np.inf

    def grad_lnlikelihood(self, y, dims=None, quiet=False):
        """
        Compute the gradient of the ln-likelihood function as a function of
        the kernel parameters.

        :param y: ``(nsamples,)``
            The list of observations at coordinates ``x`` provided to the
            :func:`compute` function.

        :param dims: (optional)
            If you only want to compute the gradient in some dimensions,
            list them here.

        :param quiet:
            If ``True`` return a gradient of zero instead of raising an
            exception when there is an invalid kernel or linear algebra
            failure. (default: ``False``)

        """
        # By default, compute the gradient in all dimensions.
        if dims is None:
            dims = np.ones(len(self.kernel), dtype=bool)

        # Make sure that the model is computed and try to recompute it if it's
        # dirty.
        if not self.computed:
            try:
                self.recompute()
            except (ValueError, LinAlgError):
                if quiet:
                    return np.zeros_like(dims, dtype=float)
                raise

        # Parse the input sample list.
        r = self._check_dimensions(y)[self.inds] - self.mean(self._x)

        # Pre-compute some factors.
        alpha = cho_solve(self._factor, r)
        Kg = self.kernel.grad(self._x[:, None], self._x[None, :])[dims]

        # Loop over dimensions and compute the gradient in each one.
        g = np.empty(len(Kg))
        for i, k in enumerate(Kg):
            d = sum(map(lambda r: np.dot(alpha, r), alpha[:, None] * k))
            d -= np.sum(np.diag(cho_solve(self._factor, k)))
            g[i] = 0.5 * d

        return g

    def predict(self, y, t):
        """
        Compute the conditional predictive distribution of the model.

        :param y: ``(nsamples,)``
            The observations to condition the model on.

        :param t: ``(ntest,)`` or ``(ntest, ndim)``
            The coordinates where the predictive distribution should be
            computed.

        Returns a tuple ``(mu, cov)`` where

        * **mu** ``(ntest,)`` is the mean of the predictive distribution, and
        * **cov** ``(ntest, ntest)`` is the predictive covariance.

        """
        if not self.computed:
            self.recompute()

        r = self._check_dimensions(y)[self.inds] - self.mean(self._x)
        xs, i = self.parse_samples(t, False)
        alpha = cho_solve(self._factor, r)

        # Compute the predictive mean.
        Kxs = self.kernel(self._x[None, :], xs[:, None])
        mu = np.dot(Kxs, alpha) + self.mean(xs)

        # Compute the predictive covariance.
        cov = self.kernel(xs[:, None], xs[None, :])
        cov -= np.dot(Kxs, cho_solve(self._factor, Kxs.T))

        return mu, cov

    def sample_conditional(self, y, t, size=1):
        """
        Draw samples from the predictive conditional distribution.

        :param y: ``(nsamples, )``
            The observations to condition the model on.

        :param t: ``(ntest, )`` or ``(ntest, ndim)``
            The coordinates where the predictive distribution should be
            computed.

        :param size: (optional)
            The number of samples to draw. (default: ``1``)

        Returns **samples** ``(N, ntest)``, a list of predictions at
        coordinates given by ``t``.

        """
        mu, cov = self.predict(y, t)
        return multivariate_gaussian_samples(cov, size, mean=mu)

    def sample(self, t, size=1):
        """
        Draw samples from the prior distribution.

        :param t: ``(ntest, )`` or ``(ntest, ndim)``
            The coordinates where the model should be sampled.

        :param size: (optional)
            The number of samples to draw. (default: ``1``)

        Returns **samples** ``(N, ntest)``, a list of predictions at
        coordinates given by ``t``.

        """
        x, _ = self.parse_samples(t, False)
        cov = self.get_matrix(x)
        return multivariate_gaussian_samples(cov, size, mean=self.mean(x))

    def get_matrix(self, t):
        """
        Get the covariance matrix at a given set of independent coordinates.

        :param t: ``(nsamples,)`` or ``(nsamples, ndim)``
            The list of samples.

        """
        r, _ = self.parse_samples(t, False)
        return self.kernel(r[:, None], r[None, :])

    def optimize(self, x, y, yerr=TINY, sort=True, dims=None, in_log=True,
                 verbose=True, **kwargs):
        """
        A simple and not terribly robust non-linear optimization algorithm for
        the kernel hyperpararmeters.

        :param x: ``(nsamples,)`` or ``(nsamples, ndim)``
            The independent coordinates of the data points.

        :param y: ``(nsamples, )``
            The observations at the coordinates ``x``.

        :param yerr: (optional) ``(nsamples,)`` or scalar
            The Gaussian uncertainties on the data points at coordinates
            ``x``. These values will be added in quadrature to the diagonal of
            the covariance matrix.

        :param sort: (optional)
            Should the samples be sorted before computing the covariance
            matrix?

        :param dims: (optional)
            If you only want to optimize over some parameters, list their
            indices here.

        :param in_log: (optional) ``(len(kernel),)``, ``(len(dims),)`` or bool
            If you want to fit the parameters in the log (this can be useful
            for parameters that shouldn't go negative) specify that here. This
            can be a single boolean---in which case it is assumed to apply to
            every dimension---or it can be an array of booleans, one for each
            dimension.

        :param verbose: (optional)
            Display the results of the call to :func:`scipy.optimize.minimize`?
            (default: ``True``)

        Returns ``(pars, results)`` where ``pars`` is the list of optimized
        parameters and ``results`` is the results object returned by
        :func:`scipy.optimize.minimize`.

        """
        self.compute(x, yerr, sort=sort)

        # By default, optimize all the hyperparameters.
        if dims is None:
            dims = np.ones(len(self.kernel), dtype=bool)
        dims = np.arange(len(self.kernel))[dims]

        # Deal with conversion functions.
        try:
            len(in_log)
        except TypeError:
            in_log = in_log * np.ones_like(dims, dtype=bool)
        else:
            if len(in_log) != len(dims):
                raise RuntimeError("Dimension list and log mask mismatch")

        # Build the conversion functions.
        conv = np.array([lambda x: x for i in range(len(dims))])
        iconv = np.array([lambda x: x for i in range(len(dims))])
        conv[in_log] = np.exp
        iconv[in_log] = np.log

        # Define the objective function and gradient.
        def nll(pars):
            for i, f, p in izip(dims, conv, pars):
                self.kernel[i] = f(p)

            ll = self.lnlikelihood(y, quiet=True)
            if not np.isfinite(ll):
                return 1e25  # The optimizers can't deal with infinities.
            return -ll

        def grad_nll(pars):
            for i, f, p in izip(dims, conv, pars):
                self.kernel[i] = f(p)
            return -self.grad_lnlikelihood(y, dims=dims, quiet=True)

        # Run the optimization.
        p0 = [f(p) for f, p in izip(iconv, self.kernel.pars[dims])]
        results = op.minimize(nll, p0, jac=grad_nll, **kwargs)

        if verbose:
            print(results.message)

        # Update the kernel.
        for i, f, p in izip(dims, conv, results.x):
            self.kernel[i] = f(p)

        return self.kernel.pars[dims], results


class _default_mean(object):

    def __init__(self, value):
        self.value = value

    def __call__(self, t):
        return self.value + np.zeros(len(t), dtype=float)
